import warnings

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.utils.deprecation import RemovedInDjango50Warning
from django.utils.inspect import func_supports_parameter

from lti_tool.types import LtiLaunch

UserModel = get_user_model()


class LtiLaunchAuthenticationBackend(ModelBackend):
    """A Django authentication backend class for LTI launch authentication.

    This backend is to be used in conjunction with the
    ``LtiLaunchAuthenticationMiddleware`` found in the middleware module of
    this package, and is used when authentication is handled by an LMS and
    passed in via LTI launch.

    By default, the ``authenticate`` method creates ``User`` objects for
    usernames that don't already exist in the database.  Subclasses can disable
    this behavior by setting the ``create_unknown_user`` attribute to
    ``False``.
    """

    # Create a User object if not already in the database?
    create_unknown_user = True

    def authenticate(self, request, lti_launch_user_id):
        """Authenticate the user based on the LTI launch.

        The username passed as ``lti_launch_user_id`` is considered trusted. Return
        the ``User`` object with the given username. Create a new ``User``
        object if ``create_unknown_user`` is ``True``.

        Return None if ``create_unknown_user`` is ``False`` and a ``User``
        object with the given username is not found in the database.
        """
        if not lti_launch_user_id:
            return
        created = False
        user = None
        username = self.clean_username(lti_launch_user_id)

        # Note that this could be accomplished in one try-except clause, but
        # instead we use get_or_create when creating unknown users since it has
        # built-in safeguards for multiple threads.
        if self.create_unknown_user:
            user, created = UserModel._default_manager.get_or_create(
                **{UserModel.USERNAME_FIELD: username}
            )
        else:
            try:
                user = UserModel._default_manager.get_by_natural_key(username)
            except UserModel.DoesNotExist:
                pass

        # RemovedInDjango50Warning: When the deprecation ends, replace with:
        #   user = self.configure_user(request, user, created=created)
        if func_supports_parameter(self.configure_user, "created"):
            user = self.configure_user(request, user, created=created)
        else:
            warnings.warn(
                f"`created=True` must be added to the signature of "
                f"{self.__class__.__qualname__}.configure_user().",
                category=RemovedInDjango50Warning,
            )
            if created:
                user = self.configure_user(request, user)
        return user if self.user_can_authenticate(user) else None

    def clean_username(self, username):
        """Clean the username.

        Perform any cleaning on the "username" prior to using it to get or
        create the user object.  Return the cleaned username.

        By default, return the username unchanged.
        """
        return username

    def configure_user(self, request, user, created=True):
        """Configure a user.

        Configure a user and return the updated user.

        By default, return the user unmodified.
        """
        if created:
            if hasattr(request, "lti_launch"):
                launch: LtiLaunch = request.lti_launch
                user.first_name = launch.user.given_name
                user.last_name = launch.user.family_name
                user.email = launch.user.email
                user.save()
                # associate the Django user with the LTI tool user
                launch.user.auth_user = user
                launch.user.save()
        return user


class AllowAllUsersLtiLaunchAuthenticationBackend(LtiLaunchAuthenticationBackend):
    def user_can_authenticate(self, user):
        return True



from django.conf import settings
from django.contrib import auth
from django.contrib.auth import load_backend
from django.core.exceptions import ImproperlyConfigured
from django.utils.deprecation import MiddlewareMixin

from lti_authentication.backends import LtiLaunchAuthenticationBackend
from lti_tool.models import LtiUser


class LtiLaunchAuthenticationMiddleware(MiddlewareMixin):
    """Middleware for utilizing LMS-provided authentication via LTI launch.

    This middleware works in conjunction with `LtiLaunchMiddleware` and
    Django's `AuthenticationMiddleware`.  The `LtiLaunchMiddleware` and
    `AuthenticationMiddleware` MUST appear before this middleware in the
    middleware list.

    If request.user is not authenticated, then this middleware attempts to
    authenticate the username from ``request.lti_launch.user.sub``.

    If authentication is successful, the user is automatically logged in to
    persist the user in the session.

    """

    force_logout_if_no_launch = True

    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, "user"):
            raise ImproperlyConfigured(
                "The Django LTI launch auth middleware requires the"
                " authentication middleware to be installed.  Edit your"
                " MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the LtiLaunchAuthenticationMiddleware class."
            )
        # LtiLaunchMiddleware is required so that request.lti_launch exists.
        if not hasattr(request, "lti_launch"):
            raise ImproperlyConfigured(
                "The Django LTI launch auth middleware requires the"
                " LTI launch middleware to be installed.  Edit your"
                " MIDDLEWARE setting to insert"
                " 'lti_tool.auth.LtiLaunchMiddleware'"
                " before the LtiLaunchAuthenticationMiddleware class."
            )

        # If the request doesn't have an LTILaunch object, then we can't
        # authenticate the user.
        if request.lti_launch.is_absent:  # pragma: no cover
            if self.force_logout_if_no_launch and request.user.is_authenticated:
                self._remove_invalid_user(request)
            return
        try:
            username = self.get_username(request)
        except LtiUser.DoesNotExist:
            # If the LTI launch user doesn't exist then remove any existing
            # authenticated remote-user, or return (leaving request.user set to
            # AnonymousUser by the AuthenticationMiddleware).
            if self.force_logout_if_no_launch and request.user.is_authenticated:
                self._remove_invalid_user(request)
            return

        # If the user is already authenticated and that user is the user we are
        # getting passed in the LTI launch, then the correct user is already
        # persisted in the session and we don't need to continue.
        if request.user.is_authenticated:
            if request.user.get_username() == self.clean_username(username, request):
                return
            else:
                # An authenticated user is associated with the request, but
                # it does not match the authorized user in the header.
                self._remove_invalid_user(request)

        # We are seeing this user for the first time in this session, attempt
        # to authenticate the user.
        user = auth.authenticate(request, lti_launch_user_id=username)
        if user:
            # User is valid.  Set request.user and persist user in the session
            # by logging the user in.
            request.user = user
            auth.login(request, user)

    def clean_username(self, username, request):
        # Allow the backend to clean the username, if the backend defines a
        # clean_username method.
        backend_str = request.session[auth.BACKEND_SESSION_KEY]
        backend = auth.load_backend(backend_str)
        try:
            username = backend.clean_username(username)  # type: ignore
        except AttributeError:  # Backend has no clean_username method.
            pass
        return username

    def get_username(self, request):
        """Get the username from the request.

        This method returns the value that will be used as the username of
        the Django user object. By default, this is the value of the
        `sub` attribute of the LTI user object.
        """
        if hasattr(settings, "LTI_AUTHENTICATION") and settings.LTI_AUTHENTICATION.get(
            "use_person_sourcedid", False
        ):
            username = request.lti_launch.get_claim(
                "https://purl.imsglobal.org/spec/lti/claim/lis"
            ).get("person_sourcedid")
        else:
            username = request.lti_launch.user.sub

        return username

    def _remove_invalid_user(self, request):
        # Remove the current authenticated user in the request which is invalid
        # but only if the user is authenticated via the LtiLaunchAuthenticationBackend.

        try:
            stored_backend = load_backend(
                request.session.get(auth.BACKEND_SESSION_KEY, "")
            )
        except ImportError:
            # backend failed to load
            auth.logout(request)
        else:
            if isinstance(stored_backend, LtiLaunchAuthenticationBackend):
                auth.logout(request)


class PersistentLtiLaunchAuthenticationMiddleware(LtiLaunchAuthenticationMiddleware):
    """Middleware for web-server provided authentication on logon pages.

    Like LtiLaunchAuthenticationMiddleware but keeps the user authenticated even if
    the ``request.META`` key is not found in the request. Useful for
    setups when the external authentication is only expected to happen
    on some "logon" URL and the rest of the application wants to use
    Django's authentication mechanism.
    """

    force_logout_if_no_launch = False
